<?php

/*

	Takeo Akiyama, Fourthgate Inc.

 */


namespace App\Controller\Component;

use Cake\Controller\Component;
use \Exception;
use Cake\Network\Socket;
use Cake\Core\Configure;
use Cake\Log\Log;

use Cake\Network\Exception\InternalErrorException;

class MpdComponent extends Component {

	//Connection & Write Errors
	const CONNECTION_NOT_OPENED   = 100;
	const CONNECTION_FAILED       = 102;
	const WRITE_FAILED            = 103;

	//MPD Errors
	const ACK_NOT_LIST            = 1;
	const ACK_ARG                 = 2;
	const ACK_PASSWORD            = 3;
	const ACK_PERMISSION          = 4;
	const ACK_UNKOWN              = 5;
	const ACK_NO_EXIST            = 50;
	const ACK_PLAYLIST_MAX        = 51;
	const ACK_SYSTEM              = 52;
	const ACK_PLAYLIST_LOAD       = 53;
	const ACK_UPDATE_ALREADY      = 54;
	const ACK_PLAYER_SYNC         = 55;
	const ACK_EXIST               = 56;
	const ACK_COMMAND_FAILED      = -100;

	//MPD Responces
	const RESPONSE_OK             = 'OK';

	private $mpdHost = 'localhost';
	private $mpdPort = '6600';
	private $mpdPassword = NULL;
	private $seekTime = 10;

	private $_connection = NULL;
	protected $_commands = array();
	protected $_output = array();
	protected $_errors = array();
	private $_current_error = array();

	private $mpd_version = '';


	public function initialize(array $config): void {
//		parent::initialize($config);
		$this->mpdHost = Configure::read('Mpd.host');
		$this->mpdPort = Configure::read('Mpd.port');
		$this->mpdPassword = Configure::read('Mpd.password');
		$this->seekTime = Configure::read('Mpdplay.seektime');
	}


	/**
	 * Connect to MPD
	 *
	 * @return bool
	 */
	private function _connect( ) {
		if ($this->_isConnected()) return TRUE;

//        Log::write( 'debug', "MpdComponent _connect: ". $this->mpdHost ."/". $this->mpdPort );
        $connection = fsockopen( $this->mpdHost, $this->mpdPort, $errn, $errs, 4 );
//        Log::write( 'debug', "albums_id s: ". print_r( $connection, true ) );
        if ($connection) {
            $this->_connection = $connection;
            while ( ! feof( $this->_connection ) ) {
                $line = fgets($this->_connection);
                $tmps = preg_match('/^OK MPD (.*?)$/', $line, $matches);
                if ( $tmps == 1 ) {
                    $this->mpd_version = $matches[1];
                }
//                Log::write( 'debug', "_connect line: ". $line );
                if (trim(substr($line, 0, 2)) == self::RESPONSE_OK) {
                    break;
                }
            }
            if (!is_null($this->mpdPassword)) {
                if ($this->runCommand('password', $this->mpdPassword) === self::ACK_PASSWORD) {
                    throw new InternalErrorException( );
                }
            }
            return true;
        } else {
//            Log::write( 'debug', "InternalErrorException: ". print_r( $connection, true ) );
            throw new InternalErrorException( );
        }
        return false;
	}


	/**
	 * Write data to the socket
	 *
	 * @param $command string data to be sent
	 *
	 * @return bool
	 *
	 */
	private function _write($data) {
		if (!$this->_isConnected()) {
			$this->_connect();
        }
		if (!fwrite($this->_connection, $data."\r\n")) {
			throw new InternalErrorException( );
		}
		$this->_commands[] = $data;
		return TRUE;
	}


	/**
	 * Run command
	 *
	 * @param $command string a command to be executed through MPD
	 * @param $args mixed string for a single argument, array for multiple
	 * @param $parse mixed false to parse the output, int for parse style
	 *
	 * @return array of server output
	 */
	public function runCommand( $command, $args = array(), $parse = 0 ) {
		if (is_array($args)) {
			foreach($args as $arg) {
				$command.= ' "'.str_replace('"', '\"', $arg) .'"';
			}
		} elseif (!is_null($args)) {
			$command.= ' "'.str_replace('"', '\"', $args) .'"';
		}
//        Log::write( 'debug', "command string: ". $command );
		$this->_write($command);
		$output = $this->_read();
//        Log::write( 'debug', "runCommand return value: ". print_r( $output, true ) );

		$this->_output[] = array($command, $output);
		if ($output === array()) return TRUE;
		if ($parse !== false) return $this->_parseOutput($output, $parse);
		return $output;
	}


	/**
	 * Parse MPD output on a line-by-line basis
	 * creating output that is easy to work with
	 *
	 * @param $input array of input from MPD
	 * @param $style int style number,'0' for the "intelligent" sorting
	 *
	 * @return array
	 */
	private function _parseOutput($input, $style = 0) {
		if ( !is_array($input) ) $input = array($input);
//		Log::write( 'debug', '_parseOutput => '. print_r( $input, TRUE ) );

		$count = array('outputs' => -1, 'file' => -1, 'key' => 0);
		$used_keys = array();
		$output = array();
		$prev = array('key' => null, 'value' => null);
		$dirtoggle = false;
		foreach($input as $line) {
			if (is_array($line)) {
				$this->_errors[] = 'Server output not expected: '.print_r($line, true);
				continue;
			} else {
				$parts = explode(': ', $line, 2);
				if (!isset($parts[0], $parts[1])) {
					$this->_errors[] = 'Server output not expected: '.$line;
					continue;
				}
			}
			$key = trim($parts[0]);
			$value = trim($parts[1]);
			if ($style == 0) {
				switch ($key) {
					//The following has to do strictly
					//with files in the output
					case 'file':
					case 'Artist':
                    case 'ArtistSort':
					case 'Album':
                    case 'AlbumSort':
                    case 'AlbumArtist':
                    case 'AlbumArtistSort':
					case 'Title':
					case 'Track':
					case 'Name':
					case 'Genre':
					case 'Date':
					case 'Composer':
                    case 'Conductor':
					case 'Performer':
					case 'Comment':
					case 'Disc':
					case 'Id':
					case 'Pos':
					case 'Time':
                    case 'Format':
                    case 'duration':
						if ($key == 'file') {
							$count['file']++;
						}
						$output['file'][$count['file']][$key] = $value;
						break;

					//The next section is for a 'stats' call
					case 'artists':
					case 'albums':
					case 'songs':
					case 'uptime':
					case 'playtime':
					case 'db_playtime':
					case 'db_update':
						$output['stats'][$key] = $value;
						break;

					//Now for a status call:
					case 'volume':
					case 'repeat':
					case 'random':
					case 'playlistlength':
					case 'xfade':
					case 'state':
					case 'song':
					case 'songid':
					case 'time':
					case 'bitrate':
					case 'audio':
					case 'updating_db':
						$output['status'][$key] = $value;
						break;

					//Outputs
					case 'outputid':
					case 'outputname':
					case 'outputenabled':
						if ($key == 'outputid') {
							$count['outputs']++;
						}
						$output['outputs'][$count['outputs']][$key] = $value;
						if ($key == 'outputid') {
							$count['outputs']++;
						}
						break;

					//The 'playlist' case works in 2 scenarios
					//1) in a file/directory listing
					//2) in a status call
					// This is to determine if it was in a status call
					// or in a directory call.
					case 'playlist':
						if ($prev['key'] == 'random') {
							$output['status'][$key] = $value;
						} else {
							$output[$key][] = $value;
						}
						break
;
					//Now that we've covered most of the weird
					//options of output,
					//lets cover everything else!
					default:
//                        Log::write( 'debug', "key: {$key} => {$value}" );
						if (isset($used_keys[$key])) {
							$used_keys = array();
							$count['key']++;
						}
						$used_keys[$key] = true;
						//$output[$count['key']][$key] = $value;//This is rarely useful
						$output[$key][] = $value;
						break;
				}
			} elseif ($style == 1) {
				$output[$key][] = $value;
			}
			if ($key == 'directory') {
				$dirtoggle = true;
			}
			$prev['key'] = $key;
			$prev['value'] = $value;
		}
		return $output;

	}


	/**
	 * Disconnect from MPD
	 *
	 * @return bool
	 */
	private function _disconnect() {
		$this->runCommand('close');
		fclose($this->_connection);
		$this->_connection = null;
		return TRUE;
	}


	public function occError( ) {
		throw new InternalErrorException( );
	}


	/**
	 * Get the status of the MPD server
	 *
	 * @return array
	 */
	public function getStatus( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$status = $this->runCommand('status');
		$this->_disconnect( );
		if ( ! isset($status['status']) ) return FALSE;

		return $status;
	}


    /**
     * Get the status of the MPD server
     *
     * @return array
     */
    public function getVersion( ) {
        if ( ! $this->_isConnected() ) $this->_connect( );
        $this->_disconnect( );
        return $this->mpd_version;
    }

    /**
     * Get the status of the MPD server
     *
     * @return array
     */
    public function getFormatsts( ) {
        if ( ! $this->_isConnected() ) $this->_connect( );
        $this->_disconnect( );

        $tmps = preg_match('/^(\d)\.(\d*)\.(.*)$/', $this->mpd_version, $matches);
        if ( isset( $matches[2] ) && ( intval( $matches[2] ) >= 21 ) ) {
            return true;
        } else {
            return false;
        }
    }


	/**
	 * Get various statistics about the MPD server
	 *
	 * @return array
	 */
	public function getStats( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stats = $this->runCommand('stats');
//        Log::write( 'debug', "getStats: ". print_r( $stats, true ) );
		$this->_disconnect( );
		if ( ! isset($stats['stats']) ) return FALSE;

		return $stats;
	}


	/*
	 *	from Net_MPD_Common
	 */


	/**
	 * Check connection status
	 *
	 * @return bool
	 */
	private function _isConnected( ) {
		if ( ! is_resource($this->_connection) ) {
			return FALSE;
		}
		return TRUE;
	}


	/**
	 * Get the current error data
	 *
	 * @return array of error data
	 */
	public function getErrorData() {
		return $this->_current_error;
	}

	/**
	 * A method to access errors
	 *
	 * @return array
	 */
	public function getErrors() {
		return $this->_errors;
	}


	/**
	 * Read data from the socket
	 *
	 * @return array of raw output
	 *
	 */
	private function _read() {
		if (!$this->_isConnected()) {
			throw new InternalErrorException( );
		}
		$output = array();
		while (!feof($this->_connection)) {
			$line = fgets($this->_connection);
//			Log::write( 'debug', 'fgets data : '. print_r( $line, TRUE ) );
			if (preg_match('/^ACK \[(.*?)\@(.*?)\] \{(.*?)\} (.*?)$/', $line, $matches)) {
				$this->runCommand('clearerror');
				$this->_errors[] = $matches;
//                Log::write( 'debug', 'fgets data : '. print_r( $line, TRUE ) );
				$this->_current_error = array('ack' => $matches[1], 'func' => $matches[3], 'error' => $matches[4]);
				throw new InternalErrorException( );
			} elseif (trim($line) == self::RESPONSE_OK) {
				break;
			} else {
				$output[] = $line;
			}
		}
		return $output;
	}


	/**
	 * Used to discover commands that are not available
	 *
	 * @return array (null on no functions not being available)
	 */
	public function getNotCommands() {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$cmds = $this->runCommand('notcommands');
		$this->_disconnect( );
		if ( ! isset($cmds['command']) ) return array();
		return $cmds['command'];
	}


	/**
	 * Used to discover which commands are available
	 *
	 * @return array (null on no functions being available
	 */
 	public function getCommands( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$cmds = $this->runCommand('commands');
		$this->_disconnect( );
		if ( ! isset($cmds['command']) ) return array();
		return $cmds['command'];
	}


	/**
	 * Ping the MPD server to keep the connection running
	 *
	 * @return bool
	 */
	public function ping( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stat = $this->runCommand('ping');
		$this->_disconnect( );
		return $stat;
	}

	/*
	 * from Net_MPD_Database
	 */

	/**
	 * Case sensitive search for data in the database
	 *
	 * @param array $params         array('search_field' => 'search for')
	 * @param bool  $caseSensitive  True for case sensitivity, false for not [default false]
	 * @return array
	 */
	public function find($params, $caseSensitive = false) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$prms = array();
		foreach($params as $key => $value) {
			$prms[] = $key;
			$prms[] = $value;
		}
		$cmd = $caseSensitive ? 'find' : 'search';

		$out = $this->runCommand($cmd, $prms);
		$this->_disconnect( );
		if (!isset($out['file'])) return array();

		return $out['file'];
	}


	/**
	 * List all metadata of matches to the search
	 *
	 * @param string $metadata1 metadata to list
	 * @param string $metadata2 metadata field to search in, optional
	 * @param string $search    data to search for in search field,
	 *                          required if search field provided
	 * @return array
	 */
	public function getMetadata($metadata1, $metadata2 = null, $search = null) {
		//Make sure that if metadata2 is set, search is as well
		if (!is_null($metadata2)) {
			if (is_null($search)) return FALSE;
		}

		if ( ! $this->_isConnected() ) $this->_connect( );
		if (!is_null($metadata2)) {
			$out = $this->runCommand('list', array($metadata1, $metadata2, $search), 1);
		} else {
			$out = $this->runCommand('list', $metadata1, 1);
		}
		$this->_disconnect( );

		if ($metadata1 == 'filename') {
			$metadata1 = 'file';
		}
		return $out[$metadata1];
	}


	/**
	 * Lists all files and folders in the directory recursively
	 *
	 * @param $dir string directory to start in, optional
	 * @return array
	 */
	public function getAll($dir = '') {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stats = $this->runCommand('listall', $dir, 1);

		$this->_disconnect( );
		return $stats;
	}


	/**
	 * Lists all files/folders recursivly, listing any related informaiton
	 *
	 * @param $dir string directory to start in, optional
	 * @return array
	 */
	public function getAllInfo($dir = '') {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stats = $this->runCommand('listallinfo', $dir);

		$this->_disconnect( );
		return $stats;
	}


	/**
	 * Lists content of the directory
	 *
	 * @param $dir string directory to work in, optional
	 * @return array
	 */
	public function getInfo($dir = '', $disconn = true ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
//        Log::write( 'debug', "getInfo dir: ". $dir );
		$stats = $this->runCommand('lsinfo', $dir);
//        Log::write( 'debug', "getInfo result: ". print_r( $stats, true ) );

		if ( $disconn ) {
//            Log::write( 'debug', "connection colse" );
		    $this->_disconnect( );
        }
		return $stats;
	}


	/*
	 * from Net_MPD_Playback
	 */

	/**
	 * Gets the current song and related information
	 *
	 * @return array of data
	 */
	public function getCurrentSong( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$out = $this->runCommand('currentsong');
//        Log::write( 'debug', "getCurrentSong: ". print_r( $out, true ) );
		$this->_disconnect( );
		if (!isset($out['file'][0])) return FALSE;
		$rets = $out['file'][0];
		if ( isset( $out['Label'][0] ) ) $rets['Label'] = $out['Label'][0];
        if ( isset( $out['AlbumArtist'][0] ) ) $rets['AlbumArtist'] = $out['AlbumArtist'][0];
        if ( isset( $out['ArtistSort'][0] ) ) $rets['ArtistSort'] = $out['ArtistSort'][0];
        if ( isset( $out['AlbumArtistSort'][0] ) ) $rets['AlbumArtistSort'] = $out['AlbumArtistSort'][0];
		return $rets;

	}

	/**
	 * Set crossfade between songs
	 *
	 * @param $sec int, seconds to crossfade
	 * @return bool
	 */
	public function setCrossfade($sec) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('crossfade', $sec);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Continue to the next song
	 *
	 * @return bool
	 */
	public function nextSong( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('next');
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Go back to the previous song
	 *
	 * @return bool
	 */
	public function previousSong( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('previous');
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Pauses or plays the audio
	 *
	 * @return bool
	 */
	public function pause( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('pause');
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Starts playback
	 *
	 * @param $song int, song position in playlist to start playing at
	 * @return bool
	 */
	public function play($song = 0) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('play', $song);
		$this->_disconnect( );
		return TRUE;
	}


	/**
	 * Starts playback by Id
	 *
	 * @param $song int, song Id
	 * @return bool
	 */
	public function playId($song = 0) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('playid', $song);
		$this->_disconnect( );
		return TRUE;
	}


    /**
     * Sets 'single' mode on/off
     *
     * @param $mode int 0 or 1, for single or all,
     * @return bool
     */
    public function single( $mode = 0 ) {
        if ( ! $this->_isConnected() ) $this->_connect( );
        if ( $mode != 1 ) $mode = 0;
        $ret = $this->runCommand('single', array( $mode ) );
        $this->_disconnect( );
        return TRUE;
    }


	/**
	 * Sets 'random' mode on/off
	 *
	 * @param $mode int 0 or 1, for random on 1,
	 * @return bool
	 */
	public function random( $mode = 0 ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('random', (int)$mode );
		$this->_disconnect( );
		return TRUE;
	}


	/**
	 * Sets 'repeat' mode on/off
	 * @access public
	 * @param $on bool true or false, for repeat or not (respectively),
     optional
	 * @return true
	 */
	public function repeat($on = 0) {
		if ( ! $this->_isConnected() ) $this->_connect( );
        $ret = $this->runCommand('repeat', (int)$on);
		$this->_disconnect( );
		return TRUE;
	}


	/**
	 * Seek a position in a song
	 *
	 * @param $song int song position in playlist
	 * @param $time int time in seconds to seek to
	 * @return bool
	 */
	public function seek($song, $time ) {
		$time = intval( $time );
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('seek', array($song, $time));
		$this->_disconnect( );
		return TRUE;
	}


	/**
	 * Seek a position in a song
	 *
	 * @param $song int song Id
	 * @param $time int time in seconds to seek to
	 * @return bool
	 */
	public function seekId($song, $time ) {
		$time = intval( $time );
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('seekid', array($song, $time));
		$this->_disconnect( );
		return TRUE;
	}


	/**
	 * Set volume
	 *
	 * @param $vol int volume
	 * @return true
	 */
	public function setVolume($vol) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('setvol', $vol);
		$this->_disconnect( );
		return TRUE;
	}


	/**
	 * Stop playback
	 *
	 * @return bool
	 */
	public function stop() {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('stop');
		$this->_disconnect( );
		return TRUE;
	}



	/*
	 *	from Net_MPD_Playlist
	 */

	/**
	 * List playlists in specified directory
	 *
	 * @param $dir string directory path, optional
	 * @return bool true on success int on failure
	 */
	public function getPlaylists($dir = '') {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$out = $this->runCommand('lsinfo', $dir);
		$this->_disconnect( );
		if ( ! isset( $out['playlist'] ) ) return array( );
		return $out['playlist'];
	}

    /**
     * load playlists
     *
     * @param $dir string directory path, optional
     * @return bool true on success int on failure
     */
    public function loadPlaylists($dir = '') {
        if ( ! $this->_isConnected() ) $this->_connect( );
        $out = $this->runCommand( 'listplaylistinfo', [ $dir ] );

        $out2 = $this->runCommand('load', [ $dir ]);
        $this->_disconnect( );
        if ( ! isset( $out['playlist'] ) ) return array( );
        return $out['playlist'];
    }


	/**
	 * List playlist info in specified directory
	 *
	 * @param $dir string directory path, optional
	 * @return bool true on success int on failure
	 */
	public function getPlaylistDetail($dir = '') {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$out = $this->runCommand( 'listplaylistinfo', [ $dir ]);
		$this->_disconnect( );
		if ( empty( $out ) ) return array( );
		return $out;
	}

	/**
	 * Add file to playlist
	 *
	 * @param $file string filename
	 * @return bool
	 */
	public function addSong($file) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('add', $file);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Clear the playlist
	 *
	 * @return bool
	 */
	public function clear() {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('clear');
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Delete song from playlist
	 *
	 * @param $song int song position in playlist
	 * @return bool
	 */
	public function deleteSong($id) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('delete', $id);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Delete song from playlist by song Id
	 *
	 * @param $id int song Id
	 * @return bool
	 */
	public function deleteSongId($id) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('deleteid', $id);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Loads a playlist into the current playlist
	 *
	 * @param $playlist string playlist name
	 * @return bool
	 */
	public function loadPlaylist($playlist) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('load', $playlist);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Move song in the playlist
	 *
	 * @param $from int song position in the playlist
	 * @param $to int song position to move it to
	 * @return bool
	 */
	public function moveSong($from, $to) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('move', array($from, $to));
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Move song in the playlist by Id
	 *
	 * @param $from int song Id
	 * @param $to int song Id to move it to
	 * @return bool
	 */
	public function moveSongId($fromId, $toId) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('moveid', array($fromId, $toId));
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Displays metadata for songs in the playlist by position Id
	 *
	 * @param $song int song position, optional
	 * @return array of song metadata
	 */
	public function getPlaylistInfo($song = null) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$out = $this->runCommand('playlistinfo', $song, 0);
		$this->_disconnect( );
		return isset($out['file']) ? $out['file'] : array();
	}


	/**
	 * Displays metadata for songs in the playlist
	 *
	 * @param $song int song Id, optional
	 * @return array of song metadata
	 */
	public function getPlaylistInfoId($song = null) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stat = $this->runCommand('playlistid', $song);
		$this->_disconnect( );
		return $stat;
	}


	/**
	 * Get playlist changes
	 *
	 * @param $version int playlist version
	 * @param $limit boolean true to limit return
	 *               to only position and id
	 *
	 * @return array of changes
	 */
	public function getChanges($version, $limit = false) {
		$cmd = $limit ? 'plchangesposid' : 'plchanges';
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stat = $this->runCommand($cmd, $version);
		$this->_disconnect( );
		return $stat;
	}

	/**
	 * Delete a playlist
	 *
	 * @param $playlist string playlist name
	 * @return true
	 */
	public function deletePlaylist($playlist) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('rm', $playlist);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Save the playlist
	 *
	 * @param $playlist string playlist name
	 * @return bool
	 */
	public function savePlaylist($playlist) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('save', $playlist);
		$this->_disconnect( );
		return TRUE;
	}

	/**
	 * Shuffle the playlist
	 *
	 * @return true
	 */
	public function shuffle() {
		$this->runCommand('shuffle');
        $this->_disconnect( );
		return TRUE;
	}

	/**
	 * Swap song by position in the playlist
	 *
	 * @param $song1 int song position from
	 * @param $song2 int song position to
	 * @return bool
	 */
	public function swapSong($song1, $song2) {
		$this->runCommand('swap', array($song1, $song2));
        $this->_disconnect( );
		return TRUE;
	}

	/**
	 * Swaps a song with another song, by Id
	 *
	 * @param $song1 int Id of the first song
	 * @param $song2 int Id of the second song
	 * @return true
	 */
	public function swapSongId($songId1, $songId2) {
		$this->runCommand('swapid', array($songId1, $songId2));
        $this->_disconnect( );
		return TRUE;
	}


	/*
	 * from Net_MPD_Admin
	 */


	/**
    * List available audio outputs
    *
    * @return array or int on failure
    */
	public function getOutputs( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$stat = $this->runCommand('outputs');
		$this->_disconnect( );
		return $stat;
	}


	/**
    * Disables an audio output
    *
    * @param $id int output Id to disable
    * @return bool
    */
	public function disableOutput($id) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		if ($this->runCommand('disableoutput', $id) == array()) return TRUE;
		$this->_disconnect( );
		return FALSE;
	}


	/**
    * Enables an audio output
    *
    * @param $id int Id to enable
    * @return bool
    */
	public function enableOutput($id) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		if ($this->runCommand('enableoutput', $id) == array()) return TRUE;
		$this->_disconnect( );
		return FALSE;
	}


	/**
    * Kills the MPD server in a safe way, saving state if possible
    *
    * @return bool
    */
	public function kill( ) {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$r = $this->runCommand('kill');
		if ($r) @$this->disconnect();
		return TRUE;
	}


	/**
    * Updates the music database
    *
    * @param $path string path which to search for music, optional
    * @return bool
    */
	public function updateDatabase($path = '') {
		if ( ! $this->_isConnected() ) $this->_connect( );
		$this->runCommand('update', $path);
		$this->_disconnect( );
		return TRUE;
	}


    /**
     * Gets the current song and related information
     *
     * @return array of data
     */
    public function getGenres( ) {
        if ( ! $this->_isConnected() ) $this->_connect( );
        $out = $this->runCommand('list', 'genre' );
//        Log::write( 'debug', "getCurrentSong: ". print_r( $out, true ) );
        $this->_disconnect( );

        return $out;

    }



}
